/*
 * fifo-test.cc
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "fifo.h"

#define DEBUG 1
#include "debug.h"

#define THREADS 1000

typedef Fifo<int> IntFifo;

IntFifo *fifo = NULL;

void *read_fifo_thread(void *arg) {
	int i;
	int sum = 0;

	DBG_MSG(2, "Hello.");
	
	fifo->target_ref();
	while (fifo->pop(&i)) {
		DBG_MSG(3, "rcv %d", i);
		sum += i;
		usleep(rand() % 100 + 1);
	}
	fifo->target_unref();

	DBG_MSG(2, "Result: %d", sum);
	return (void*) new int(sum);
}

int main (void) {
	int i, res = 0, launched = 0, total = 0, *p = NULL;
	pthread_t thread[THREADS];

	srand(time(NULL));
	
	fifo = new IntFifo();
	
	DBG_MSG(1, "Creating threads...");
	for (i = 0; i < THREADS; i++) {
		if(pthread_create(&(thread[i]), NULL, read_fifo_thread, NULL)) {
			perror("pthread_create");
			break;
		} else {
			launched++;
		}
	}
	DBG_MSG(1, "%d threads.", launched);
	
	DBG_MSG(1, "Sending data...");
	for (i = 0; i < 1000; i++) {
		fifo->push(i);
		res += i;
		DBG_MSG(3, "snd %d\n", i);
		usleep(rand() % 100 + 1);
	}
	fifo->source_unref();
	DBG_MSG(1, "Result should be %d", res);
	
	DBG_MSG(1, "Joining threads...");
	for (i = 0; i < launched; i++) {
		pthread_join(thread[i], (void**)&p);
		total += *p;
	}
	DBG_MSG(1, "Total: %d, should be %d.", total, res * launched);

	DBG_MSG(1, "Removing fifo...");
	delete fifo;

	DBG_MSG(1, "End.");
	return total == res * launched ? EXIT_SUCCESS : EXIT_FAILURE;

}

