/*
 * screen_window.c -- OpenGL screen window
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <GL/glut.h>						// hlavickovy soubor funkci GLUTu
#include <list>

#include "screen-gl.h"

#define DEBUG 5
#include "debug.h"

//#define DEG(x)				((float)(x) / (2.0 * 3.141592))
#define RAD(x)					((float)(x) * 3.141592 / 180.0)

//static int screen_window_created = 0;

GlutMaster *ScreenWindow::glut_master = 0;

/*****************************************************************************
 * Init
 */
ScreenWindow::ScreenWindow(int *argc, char *argv[]) {
	DBG_MSG(5, "ScreenWindow::ScreenWindow");

	/* members init */
	title = "Krtek (%0.1f FPS)";
	scene_width = 74, scene_height = 20;
	scene_border = 10000;

	angle_x = 0;
	angle_y = 250;
	angle_z = 0;
	pos_x = 0;
	pos_y = 5;
	pos_z = 0;
	move_step = 0.25;
	angle_step = 2;

	fps_old_time = 0;
	fps_update_rate = 3;
	
	mouse_move = 0;
	
	/* only one window */
	/*
	if (screen_window_created) {
		fprintf(stderr, "Critical Internal Error: ScreenWindow can be created only once !!!\n");
		abort();
	}
	screen_window_created = 1;
	*/

	/* GLUT Init */
	if (!glut_master) {
		glut_master = new GlutMaster(argc, argv);
		glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
	}
	glut_master->CallGlutCreateWindow(title, this);
	//glut_master->EnableIdleFunction();
	glut_master->SetIdleToCurrentWindow();

	/* glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); */
	//glutReshapeWindow(640/2, 480/2);
	//glutSetCursor(GLUT_CURSOR_NONE);
	
	/* GL Init*/
	glEnable(GL_DITHER);
	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_POLYGON_SMOOTH);
	glDisable(GL_LINE_STIPPLE);
	glDisable(GL_POLYGON_STIPPLE);

	/* Light init */
	light_position[0] = -100.0;
	light_position[1] = -500.0;
	light_position[2] =  1000.0;
	light_position[3] =  0.0;
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);      // nasstaveni pozice svetla
	glEnable(GL_LIGHTING);                      // globalni povoleni stinovani
	glEnable(GL_LIGHT0);                        // povoleni prvniho svetla
}


/*****************************************************************************
 * CleanUp
 */
ScreenWindow::~ScreenWindow(void) {
	DBG_MSG(5, "ScreenWindow::~ScreenWindow");
	
}

/*****************************************************************************
 * exec
 */
void ScreenWindow::exec (void) {
	DBG_MSG(5, "ScreenWindow::exec");
	/* Main */
	glut_master->CallGlutMainLoop();
//	glutMainLoop();	// nekonecna smycka, kde se volaji zaregistrovane funkce
}

/*****************************************************************************
 * CallBackResizeFunc -- glut callback
 */
void ScreenWindow::CallBackReshapeFunc(int width, int height) {
	DBG_MSG(5, "ScreenWindow::CallBackReshapeFunc");
	glViewport(0, 0, width, height);		// viditelna oblast pres cele okno
	glMatrixMode(GL_PROJECTION);			// zacatek modifikace projekcni matice
	
	win_width = width;
	win_height = height;
	glutPostRedisplay();
}

/*****************************************************************************
 * CallBackDisplayFunc -- glut callback
 */
void ScreenWindow::CallBackDisplayFunc(void) {
	static int border_dl = -1;
	DBG_MSG(6, "ScreenWindow::CallBackDisplayFunc");
	//GLfloat light_position[] = {-500.0f, -500.0f, 500.0f, 1.0f}; // pozice svetla: jednicka na konci->bodove svetlo

	count_and_update_fps();
	
	/* create object */
	if (border_dl == -1) {
		DBG_MSG(4, "ScreenWindow::CallBackDisplayFunc: Creating DisplayList...");
		border_dl = glGenLists(1);
		glNewList(border_dl, GL_COMPILE);

		/* konvicka */
		gen_material(0.6, 0.6, 1.0);
		glPushMatrix();
		glTranslatef(10, 1.8, 10);
		glutSolidTeapot(2);
		glPopMatrix();
		
		
		/*
		glPointSize(5);
		glBegin(GL_POINTS);
		gen_material(1, 1, 1);
		glVertex3f(0.0, 0.0, 0.0);
		gen_material(1, 0, 0);
		glVertex3f(10.0, 0.0, 0.0);
		gen_material(0, 1, 0);
		glVertex3f(0.0, 10.0, 0.0);
		gen_material(0, 0, 1);
		glVertex3f(0.0, 0.0, 10.0);
		glEnd();
		*/

		/* gray strip around
		 *    a  b    c  d
		 *  1 +  +    +  +
		 *       |    |
		 *  2 +--+----+--+
		 *       |    |
		 *  3 +--0----+--+
		 *       |    |
		 *  4 +  +    +  +
		 */
		gen_material(0.05, 0.4, 0.0);
		glBegin(GL_TRIANGLE_FAN);
		/* b 3 */ glVertex3f(0,					0, 0);
		/* a 3 */ glVertex3f(- scene_border,			0, 0);
		/* a 4 */ glVertex3f(- scene_border,			0, - scene_border);
		/* b 4 */ glVertex3f(0,					0, - scene_border);
		glEnd();
		glBegin(GL_TRIANGLE_FAN);
		/* c 3 */ glVertex3f(scene_width,			0, 0);
		/* c 4 */ glVertex3f(scene_width,			0, - scene_border);
		/* d 4 */ glVertex3f(scene_width + scene_border,	0, - scene_border);
		/* d 3 */ glVertex3f(scene_width + scene_border,	0, 0);
		glEnd();
		glBegin(GL_TRIANGLE_FAN);
		/* c 2 */ glVertex3f(scene_width,			0, scene_height);
		/* d 2 */ glVertex3f(scene_width + scene_border,	0, scene_height);
		/* d 1 */ glVertex3f(scene_width + scene_border,	0, scene_height + scene_border);
		/* c 1 */ glVertex3f(scene_width,			0, scene_height + scene_border);
		glEnd();
		glBegin(GL_TRIANGLE_FAN);
		/* b 2 */ glVertex3f(0,					0, scene_height);
		/* b 1 */ glVertex3f(0,					0, scene_height + scene_border);
		/* a 1 */ glVertex3f(- scene_border,			0, scene_height + scene_border);
		/* a 2 */ glVertex3f(- scene_border,			0, scene_height);
		glEnd();
		glBegin(GL_QUADS);
		/* b 2 */ glVertex3f(0,					0, scene_height);
		/* a 2 */ glVertex3f(- scene_border,			0, scene_height);
		/* a 3 */ glVertex3f(- scene_border,			0, 0);
		/* b 3 */ glVertex3f(0,					0, 0);
		glEnd();
		glBegin(GL_QUADS);
		/* b 3 */ glVertex3f(0,					0, 0);
		/* b 4 */ glVertex3f(0,					0, - scene_border);
		/* c 4 */ glVertex3f(scene_width,			0, - scene_border);
		/* c 3 */ glVertex3f(scene_width,			0, 0);
		glEnd();
		glBegin(GL_QUADS);
		/* c 3 */ glVertex3f(scene_width,			0, 0);
		/* d 3 */ glVertex3f(scene_width + scene_border,	0, 0);
		/* d 2 */ glVertex3f(scene_width + scene_border,	0, scene_height);
		/* c 2 */ glVertex3f(scene_width,			0, scene_height);
		glEnd();
		glBegin(GL_QUADS);
		/* c 2 */ glVertex3f(scene_width,			0, scene_height);
		/* c 1 */ glVertex3f(scene_width,			0, scene_height + scene_border);
		/* b 1 */ glVertex3f(0,					0, scene_height + scene_border);
		/* b 2 */ glVertex3f(0,					0, scene_height);
		glEnd();

	
		gen_material(0.2, 0.6, 0.0);
		glBegin(GL_POLYGON);
		glVertex3f(0, 0, 0);
		glVertex3f(0, 0, scene_height);
		glVertex3f(scene_width, 0, scene_height);
		glVertex3f(scene_width, 0, 0);
		glEnd();
		
		glEndList();
		DBG_MSG(4, "ScreenWindow::CallBackDisplayFunc: Done.");
	}
	
	/* Vymazani */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//glOrtho(-win_width/2, win_width/2, -win_height/2, win_height/2, 0, 4000);
	gluPerspective(40, ((float)win_width)/((float)win_height), 0.1, 10000);
	glClearColor(0.6, 0.6, 1.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	/* Camera */
	//glPushMatrix();
	//glScalef(1.0, 1.0, 1.0);
	
	/* Render objects */
	glMatrixMode(GL_PROJECTION);
	glRotatef(-(angle_x - 15), 1.0, 0.0, 0.0);
	glRotatef(-angle_y, 0.0, 1.0, 0.0);
	glRotatef(-angle_z, 0.0, 0.0, 1.0);
	glTranslatef(-pos_x, -pos_y, pos_z);
	glCallList(border_dl);
	
	/* Ok, to je vse */
	glFlush();
	glutSwapBuffers();

	/* jeste prozradime pozici kamery */
	DBG_MSG(4, "ScreenWindow::CallBackDisplayFunc: pos = [ %0.1f, %0.1f, %0.1f ], angle = [ %0.1f, %0.1f, %0.1f ]",
			pos_x, pos_y, pos_z, angle_x, angle_y, angle_z);
}

/*****************************************************************************
 * CallBackVisibleFunc -- glut callback
 */
void ScreenWindow::CallBackVisibleFunc (int vis) {
	DBG_MSG(5, "ScreenWindow::CallBackVisibleFunc");
}

/*****************************************************************************
 * CallBackIdleFunc -- glut callback
 */
void ScreenWindow::CallBackIdleFunc (void) {
	DBG_MSG(6, "ScreenWindow::CallBackIdleFunc: pos: [ %lf, %lf, %lf ]", pos_x, pos_y, pos_z);
	pos_z += move_step / 4.0 * cos(RAD(angle_y)) * cos(RAD(angle_x));
	pos_x -= move_step / 4.0 * sin(RAD(angle_y)) * cos(RAD(angle_x));
	pos_y += move_step / 4.0 * sin(RAD(angle_x));
	glutPostRedisplay();
	usleep(45000);
}
/*****************************************************************************
 * CallBackKeyboardFunc -- glut callback
 * there is no keyboard
 */
void ScreenWindow::CallBackKeyboardFunc(unsigned char key, int x, int y) {
	DBG_MSG(5, "ScreenWindow::CallBackKeyboardFunc: key = %c (%d); x = %d; y = %d", key, key, x, y);
	switch (key) {
		
		/* exit */
		case 27:
		case 'q':
		case 'Q':
			exit(0);
			break;

		case 'k':
			glDisable(GL_LIGHTING);
			break;
		case 'l':
			glEnable(GL_LIGHTING);
			break;

		/* Go forward */
		case 'w':
		case 'W':
			pos_z += move_step * cos(RAD(angle_y)) * cos(RAD(angle_x));
			pos_x -= move_step * sin(RAD(angle_y)) * cos(RAD(angle_x));
			pos_y += move_step * sin(RAD(angle_x));
			break;
			
		/* Go backward */
		case 's':
		case 'S':
			pos_z -= move_step * cos(RAD(angle_y)) * cos(RAD(angle_x));
			pos_x += move_step * sin(RAD(angle_y)) * cos(RAD(angle_x));
			pos_y -= move_step * sin(RAD(angle_x));
			break;
		
		/* Go left */
		case 'a':
		case 'A':
			pos_z += move_step * cos(RAD(angle_y + 90));
			pos_x -= move_step * sin(RAD(angle_y + 90));
			break;
		
		/* Go right */
		case 'd':
		case 'D':
			pos_z += move_step * cos(RAD(angle_y - 90));
			pos_x -= move_step * sin(RAD(angle_y - 90));
			break;

		/* Go Up only */
		case '+':
		case '=':
			pos_y += move_step;
			break;

		/* Go Down only */
		case '-':
		case '_':
			pos_y -= move_step;
			break;

		/* positions */
		case '1': // prehled z dalky
			pos_x   = scene_width / 2;
			pos_y   =   50.0;
			pos_z   = scene_height * 3.0;
			angle_x =  -20.0;
			angle_y =  180.0;
			angle_z =    0.0;
			break;
		case '2': // pohled do hrdla
			pos_x   =   13.7; pos_y   =    3.4; pos_z   =   -10.0;
			angle_x =  -35.0; angle_y = -270.0; angle_z =     0.0;
			break;
	}
	if (pos_y <= 0.2) {
		pos_y = 0.2;
	}
	glutPostRedisplay();
}

/*****************************************************************************
 * CallBackSpecialFunc -- glut callback
 */
void ScreenWindow::CallBackSpecialFunc(int key, int x, int y) {
	DBG_MSG(5, "ScreenWindow::CallBackSpecialFunc: key = %d; x = %d; y = %d", key, x, y);
	switch (key) {
		case GLUT_KEY_LEFT:
			angle_y += angle_step;
			break;
		case GLUT_KEY_RIGHT:
			angle_y -= angle_step;
			break;
		case GLUT_KEY_UP:
			angle_x += angle_step;
			break;
		case GLUT_KEY_DOWN:
			angle_x -= angle_step;
			break;
	}
	DBG_MSG(4, "ScreenWindow::CallBackSpecialFunc: angle_x = %f; angle_y = %f", angle_x, angle_y);
	glutPostRedisplay();
}

void ScreenWindow::CallBackMotionFunc(int x, int y) {
	DBG_MSG(5, "ScreenWindow::CallBackMotionFunc: x = %d; y = %d", x, y);
	if (mouse_move) {
		angle_y -= (x - mouse_last_x);
		angle_x -= (y - mouse_last_y);
		mouse_last_x = x;
		mouse_last_y = y;
		glutPostRedisplay();
	}
}

void ScreenWindow::CallBackMouseFunc(int button, int state, int x, int y) {
	DBG_MSG(5, "ScreenWindow::CallBackMouseFunc: btn = %d, state = %d, x = %d; y = %d", button, state, x, y);
	mouse_move = (state == 0);
	mouse_last_x = x;
	mouse_last_y = y;
}

/*****************************************************************************
 * gen_material
 * nastaveni nejakeho rozumneho materialu dane barvy
 */
void ScreenWindow::gen_material(GLfloat r, GLfloat g, GLfloat b) {
	const float ambient = 0.6;
	//const float specular = 3;
	
	GLfloat materialAmbient[]={r*ambient, g*ambient, b*ambient, 1.0f};
	GLfloat materialDiffuse[]={r, g, b, 1.0f};
	/*
	GLfloat materialSpecular[]={r*specular, g*specular, b*specular, 1.0f};
	GLfloat materialShininess[]={150.0f};
	// */
	
	glColor3f(r, g, b);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, materialAmbient);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, materialDiffuse);
	/*
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, materialSpecular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, materialShininess);
	// */
}

/*****************************************************************************
 * count_and_update_fps
 * pocitadlo snimku za sekundu, pise hodnotu do titulku
 */
void ScreenWindow::count_and_update_fps(void) {
	/* count FPS */
	int now = time(NULL);	// TODO neco presnejsiho
	if (now > fps_old_time + fps_update_rate || fps_old_time == 0) {
		float fps = fps_counter / (now - fps_old_time);
		fps_old_time = now;
		fps_counter = 0;
		
		DBG_MSG(4, "ScreenWindow::count_and_update_fps: fps = %0.3f", fps);

		char str[256];
		sprintf(str, title, (float)fps);
		glutSetWindowTitle(str);
	}
	fps_counter += 1;
}

