/*
 * screen_window.h -- OpenGL screen window
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#ifndef JK_SCREEN_OPENGL_WINDOW_H
#define JK_SCREEN_OPENGL_WINDOW_H

#include <time.h>

#include "glut_window.h"
#include "glut_master.h"

class ScreenWindow : public GlutWindow {
	public:
		ScreenWindow(int *argc, char *argv[]);
		~ScreenWindow(void);
		void exec(void);

	private:
		static GlutMaster *glut_master;
		char *title;	// window title

		/* fps counter */
		time_t fps_old_time;		// last fps view
		float fps_counter;		// temporary value
		float fps_update_rate;		// seconds between fps updates

		/* rotation */
		int win_width, win_height;
		int scene_width, scene_height;
		int scene_border;
		float angle_x, angle_y, angle_z, pos_x, pos_y, pos_z;
		float move_step, angle_step;
		float light_position[4];

		/* mouse movement */
		int mouse_move;
		int mouse_last_x, mouse_last_y;

		void CallBackReshapeFunc (int width, int height);
		void CallBackDisplayFunc (void);
		void CallBackVisibleFunc (int vis);
		void CallBackIdleFunc (void);
		void CallBackKeyboardFunc (unsigned char key, int x, int y);
		void CallBackSpecialFunc(int key, int x, int y);
		void CallBackMotionFunc(int x, int y);
		void CallBackMouseFunc(int button, int state, int x, int y);
		
		void count_and_update_fps(void);
		void gen_material(GLfloat r, GLfloat g, GLfloat b);
};

#endif

